/*
 * ReDirector
 * 
 * Created Oct.2002
 * 
 * Copyright (c) 2002, 2003, 2005 Johann Burkard (<mailto:jb@eaio.com>)
 * <http://eaio.com>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */
package com.eaio.j2ee.www.redirector;

import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import nanoxml.XMLElement;

/**
 * A servlet that redirects users based on data from an XML file. The file
 * should have the following structure:
 * <pre>
 * &lt;links&gt;
 *  ...
 *  &lt;link&gt;
 *   &lt;id&gt;Link id&lt;/id&gt;
 *   &lt;href&gt;Link URL&lt;/href&gt;
 *  &lt;/link&gt;
 *  ...
 * &lt;links&gt;
 * </pre>
 * <p> 
 * Invoke the servlet with <code>t</code> as the parameter containing the key,
 * for example:
 * <p>
 * <code>servlet?t=RAH</code>
 * <p>
 * The t parameter is not case sensitive.
 * <p>
 * The servlets needs an init parameter called "links" that specifies where
 * to find the XML file containing the links.
 * <p>
 * CHANGES:
 * <ul>
 * <li>26.Oct.2005: Revised and moved to Grabbag project.
 * <li>01.May.2003: Minor performance tweaks.</li>
 * <li>19.Jan.2003: Added the ability to specify a host to redirect to.</li>
 * <li>03.Dec.2002: Minor changes.</li>
 * <li>28.Oct.2002: Added reload interval</li>
 * </ul>
 *  
 * @author <a href="mailto:jb@eaio.de">Johann Burkard</a>
 * @version $Id$
 */
public class ReDirector extends HttpServlet {

    /**
     * The serial version UID.
     */
    static final long serialVersionUID = -8851121474619027311L;

    /**
     * The "hostname" parameter name.
     */
    public static final String PARAMETER_HOSTNAME = "hostname";

    /**
     * The "interval" parameter name.
     */
    public static final String PARAMETER_INTERVAL = "interval";

    /**
     * The "links" parameter name.
     */
    public static final String PARAMETER_LINKS = "links";

    private Map linkMap;
    private String hostName;

    private long lastUpdate;
    private long interval;

    // interval to re-read configuration in ms
    // ~1 hour default interval

    /**
     * No-argument constructor. Does nothing.
     */
    public ReDirector() {}

    /**
     * Initializes the servlet.
     * 
     * @param config the Servlet's configuration
     * @throws ServletException
     */
    public void init(ServletConfig config) throws ServletException {
        super.init(config);
        createLinks();
    }

    /**
     * Reads and parses the configuration and
     * the XML file.
     */
    private void createLinks() {
        try {
            long itvl = Long.parseLong(getServletConfig().getInitParameter(
                    PARAMETER_INTERVAL));
            interval = (itvl < 0 ? itvl * -1 : itvl);
        }
        catch (NumberFormatException ex) {
            log(getClass().getName()
                    + ": createLinks() - Could not parse the '"
                    + PARAMETER_INTERVAL + "' parameter (should be a number), "
                    + "using default value of " + interval + " ms", ex);
            interval = 2 << 22;
        }

        hostName = getServletConfig().getInitParameter(PARAMETER_HOSTNAME);

        long start = System.currentTimeMillis();
        XMLElement xml, link, inner;
        xml = new XMLElement();

        InputStreamReader reader = null;
        String name, id, href;
        try {
            reader = new InputStreamReader(
                    getServletConfig().getServletContext().getResourceAsStream(
                            getServletConfig().getInitParameter(PARAMETER_LINKS)));
            xml.parseFromReader(reader);
        }
        catch (IOException ex) {
            log(getClass().getName()
                    + ": createLinks() - Could not parse file '"
                    + getServletConfig().getInitParameter(PARAMETER_LINKS)
                    + "', cannot continue", ex);
            return;
        }
        finally {
            if (reader != null) {
                try {
                    reader.close();
                }
                catch (IOException ex) {}
            }
        }

        linkMap = new HashMap(89);
        Enumeration links, linkParameters;
        links = xml.enumerateChildren();
        while (links.hasMoreElements()) {
            link = (XMLElement) links.nextElement();
            linkParameters = link.enumerateChildren();
            id = href = null;
            while (linkParameters.hasMoreElements()) {
                inner = (XMLElement) linkParameters.nextElement();
                name = inner.getName();
                if (name.equalsIgnoreCase("id")) {
                    id = inner.getContent().toLowerCase();
                }
                if (name.equalsIgnoreCase("href")) {
                    href = inner.getContent();
                }
            }
            if ((id != null) && (href != null)) { // id and href is specified
                linkMap.put(id, href);
            }
        }
        lastUpdate = System.currentTimeMillis();
        log(getClass().getName() + ": createLinks() - Parsed file '"
                + getServletConfig().getInitParameter(PARAMETER_LINKS)
                + "' with " + linkMap.size() + " links in "
                + (lastUpdate - start) + " ms, reloading this file every "
                + interval + " ms (" + (interval / 1000)
                + " s), host name is '" + hostName + "'");
    }

    /**
     * Clears pointers to fields.
     */
    public void destroy() {
        linkMap.clear();
        linkMap = null;
        super.destroy();
    }

    /**
     * Forwards requests to either external
     * URLs or internal pages.
     * 
     * @param request the request
     * @param response the response
     * @throws ServletException
     * @throws IOException
     */
    public void doGet(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {

        synchronized (this) {
            if ((System.currentTimeMillis() - lastUpdate) > interval) {
                // recreate the links
                createLinks();
            }
        }

        String target = request.getParameter("t");
        if ((target == null)
                || (!linkMap.containsKey((target = target.toLowerCase())))) {
            response.sendError(HttpServletResponse.SC_NOT_FOUND);
            return;
        }

        target = (String) linkMap.get(target);

        if (target.indexOf("://") == -1) {
            // internal link
            if (hostName == null) {
                response.sendRedirect(response.encodeRedirectURL(target));
            }
            else {
                response.sendRedirect(response.encodeRedirectURL(new StringBuffer(
                        hostName.length() + target.length()).append(hostName).append(
                        target).toString()));
            }
        }
        else {
            // external link
            response.setContentType("text/html; charset=ISO-8859-1");
            response.setContentLength(13);
            response.addHeader(
                    "Refresh",
                    new StringBuffer(target.length() + 5).append("0;url=").append(
                            target).toString());
            ServletOutputStream out = response.getOutputStream();
            out.print("<html></html>");
        }
    }

    /**
     * Forwards request to either external
     * URLs or internal pages.
     * 
     * @param request the request
     * @param response the response
     * @throws ServletException
     * @throws IOException
     */
    public void doPost(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        doGet(request, response);
    }

    /**
     * Return servlet information.
     * 
     * @return String
     */
    public String getServletInfo() {
        return "ReDirector - XML-based redirection servlet";
    }

    /**
     * Return the time the servlet
     * reloaded the XML file last time.
     * 
     * @return the timestamp
     */
    public long getLastModified(HttpServletRequest request) {
        return lastUpdate;
    }

}